import location,sys,e32,struct,socket,sysinfo,inbox,telephone,logs,time;

class Protocol:

    def writeCreate(self, type, userId, username):
        msg = "c"
        msg += self.intTo4Bytes(type)
        msg += self.intTo4Bytes(userId)
        msg += self.intTo4Bytes(len(username))
        msg += username
        return msg

    def writeSubscribe(self, type, userId, componentId):
        return self.writeGeneric("s", type, userId, componentId)

    def writeGeneric(self, command, type, userId, componentId):
        msg = command
        msg += self.intTo4Bytes(type)
        msg += self.intTo4Bytes(userId)
        msg += self.intTo4Bytes(componentId)
        return msg

    def writePushData(self, componentId, data):
        msg = "p"
        msg += self.intTo4Bytes(componentId)
        msg += self.writeGenericData(data)
        return msg

    def writeGenericData(self, data):
        msg = self.intTo4Bytes(len(data))
        msg += data
        return msg

    def switchRead(self, msg):
        if(msg[0] == "d"):
            return self.readGeneric(msg)
        elif (msg[0] == "a"):
            return self.readAck(msg)
        elif (msg[0] == "u"):
            return self.readGenericData(msg)
           
    def readAck(self, msg):
        command = msg[1]
        status = ord(msg[2])
        length = self.BytesToInt(msg[3:7])
        data = msg[7:7+length]
        return command, status, data

    def readGeneric(self, msg):
        type = self.BytesToInt(msg[1:5])
        user = self.BytesToInt(msg[5:9])
        componentId = self.BytesToInt(msg[9:13])
        return type, user, componentId

    def readGenericData(self, msg):
        type, user, compId = self.readGeneric(msg)
        length = self.BytesToInt(msg[13:17])
        data = msg[17:17+length]
        return type, user, compId, data

    def BytesToInt(self, bytes):
        value = ord(bytes[0]) * 256 * 256 * 256
        value += ord(bytes[1]) * 256 * 256
        value += ord(bytes[2]) * 256
        value += ord(bytes[3])
        return value

    def intTo4Bytes(self, intg):
        int1 = chr(intg % 256)
        int2 = chr((intg / 256) % 256)
        int3 = chr(((intg / 256) / 256 ) % 256)
        int4 = chr((((intg / 256) / 256 ) / 256))
        return int4 + int3 + int2 + int1
        
class Everything:
    
    def __init__(self):
        self.reset();
        self.old_cell = None;
        self.time = 0;
        self.transition = 0;
        self.gsmSock = None;
        self.callSock = None;
        self.p = Protocol();
        self.log = PhoneLog();
        self.phone_line = PhoneListener(self.log);
        self.sms = MessageListener();
        telephone.call_state(self.phone_line); # phone_line.__call__() will now be called whenever the state changes
        self.ibox = inbox.Inbox();
        self.ibox.bind(self.sms); # sms.__call__() will now be called whenever a message arrives
        self.new_calls = [];
        self.new_sms = [];
        self.new_mail = [];
        self.gsmSock = self.connectToBetelgeuse(bb_type = 0x204, bb_user_id = 1, bb_username = "gsm_sender");
        self.callSock = self.connectToBetelgeuse(bb_type = 0x205, bb_user_id = 1, bb_username = "call_sender");
        
    def connectToBetelgeuse(self, bb_type, bb_user_id, bb_username):
        sock = None;
        # connect to bb
        while (sock == None):
          try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM);
          except socket.error, msg:
            sock = None;
            e32.ao_sleep(5);
            continue;
          try:
            sock.connect(('127.0.0.1', 10004));
          except socket.error, msg:
            sock.close();
            sock = None;
            e32.ao_sleep(5);
            
        # Connection established and working
        print "Socket connected to 10004";
        msg = self.p.writeCreate(bb_type, bb_user_id, bb_username);
        sock.send(msg);
        print "Sent create to 10004";
        type, user, componentId = self.p.switchRead(sock.recv(255));
        print "Received created message with type=" + str(type) + " user=" + str(user) + " cid=" + str(componentId);
        return sock;
    
    def sendGsm(self):
        msg = self.p.writePushData(256*256*256*256-1, str(self.gsmStr()));
        try:
            self.gsmSock.send(msg);
        except socket.error, msg:
            self.gsmSock.close();
            self.gsmSock = None;
            e32.ao_sleep(5);
            self.gsmSock.connect(('127.0.0.1', 10004));
    
    def sendCalls(self):
        msg = self.p.writePushData(256*256*256*256-1, str(self.callStr()));
        try:
            self.callSock.send(msg);
        except socket.error, msg:
            self.callSock.close();
            self.callSock = None;
            e32.ao_sleep(5);
            self.callSock.connect(('127.0.0.1', 10004));        
        
    def update(self):
        ## GSM Sender
        new_cell = location.gsm_location();
        if new_cell is not None:
            (self.country_code, self.network_code, self.area_code, self.cell_id) = new_cell
        else:
            self.reset();
        self.ss = sysinfo.signal_dbm();
        self.bars = sysinfo.signal_bars();
        self.battery = sysinfo.battery();
        self.profile = sysinfo.active_profile();
        self.user_inactive_time = e32.inactivity(); 
        self.free_ram = sysinfo.free_ram();
        self.free_drivespace = sysinfo.free_drivespace();
        
        if self.old_cell is None:
            self.old_cell = self.cell_id;
        elif(self.old_cell != self.cell_id):
            self.transition = 1;
            self.old_cell = self.cell_id;
        else:
            self.transition = 0;
        
        ## Call Sender
        self.inbox_size = len(inbox.Inbox().sms_messages());
        # The PhoneListener, which is used below, waits for the line to clear before
        # a log entry is requested. Otherwise we could get an incomplete log entry
        # in the middle of a call and then we wouldn't get the length of the call.
        self.new_calls = self.phone_line.new_calls;
        self.phone_line.new_calls = []; # Reset PhoneListener's list of calls
        self.new_sms = self.log.getNewEvents('sms');
        self.new_mail = self.log.getNewEvents('email');
        self.sms.update();
            
    def reset(self):
        self.country_code = 0;
        self.network_code = 0;
        self.area_code = 0;
        self.cell_id = 0;
        self.ss = 0;
        self.bars = 0;
        self.battery = sysinfo.battery();
        self.profile = sysinfo.active_profile();
        self.user_inactive_time = 0;
        self.free_ram = 0;
        self.free_drivespace = 0;
        self.inbox_size = 0;
        self.transition = 0;
        self.new_calls = [];
        self.new_sms = [];
        self.new_mail = [];
    
    # Returns a verbose and user readable string representation of the current phone data.
    def formattedString(self):
        text = "Transition: " + str(self.transition) + "\n"
        text += "Country: " + str(self.country_code) + ", "
        text += "Network: " + str(self.network_code) + ", "
        text += "Area: " + str(self.area_code) + ", "
        text += "CellID: " + str(self.cell_id) + "\n"
        text += "Signal strength: " + str(self.ss) + "\n"
        text += "Signal bars: " + str(self.bars) + "\n"
        text += "Battery: " + str(self.battery) + "\n"
        text += "Profile: " + str(self.profile) + "\n"
        text += "User inactive for: " + str(self.user_inactive_time) + " s\n"
        text += "Free RAM: " + str(self.free_ram) + " B\n"
        text += "Messages in inbox: " + str(self.inbox_size) + "\n"
        text += "Unread messages: " + str(len(self.sms.unread_messages)) + "\n"
        text += "Phone status: " + str(self.phone_line) + "\n"
        text += "Latest calls" + " (" + str(len(self.new_calls)) + "): "
        if (self.new_calls):
            for call in self.new_calls:
                text += str(Phonecall(call).formattedString())
        text += "\nLatest messages" + " (" + str(len(self.new_sms)) + "): "
        if (self.new_sms):
            for sms in self.new_sms:
                text += str(Message(sms).formattedString())
        text += "\nRecently read messages" + " (" + str(len(self.sms.newly_read_messages)) + "): "
        if (self.sms.newly_read_messages):
            for sms_id in self.sms.newly_read_messages:
                text += str(Message(None, sms_id).formattedString())
        return text
    
    # Returns a concise string representation of the current phone data.
    # This format is used when transmitting data to Betelgeuse.
    def __str__(self):
        s = self.gsmStr() + '\t' + self.callStr()
        return s
    
    def gsmStr(self):
        s = str(self.transition) + '\t' + str(self.country_code) + '\t' + str(self.network_code) + '\t' + str(self.area_code) + '\t' + str(self.cell_id) + '\t' + str(self.ss) +'\t' + str(self.bars) + '\t' + str(self.battery) + '\t' + str(self.profile)
        s += '\t' + str(self.user_inactive_time) + '\t' + str(self.free_ram)
        return s
    
    def callStr(self):
        s = str(self.inbox_size) + '\t' + str(len(self.sms.unread_messages)) + '\t' + str(self.phone_line) + '\t'
        
        s += str(len(self.new_calls)) # number of new calls
        if self.new_calls:
            for call in self.new_calls:
                s += '\t' + str(Phonecall(call))
                
        s += '\t' + str(len(self.new_sms))   # number of new sms
        if self.new_sms:
            for sms in self.new_sms:
                s += '\t' + str(Message(sms))
                
        s += '\t' + str(len(self.sms.newly_read_messages))
        if (self.sms.newly_read_messages):   # number of newly read sms
            for sms_id in self.sms.newly_read_messages:
                s += '\n' + str(Message(None, sms_id))
        s += '\t'
        
        return s

#
# Class PhoneLog makes the logs module more convenient to use in this context
#
class PhoneLog:
    def __init__(self):
        now = time.time() + time.altzone
        self.last_read_times = {'call': now, 'sms': now, 'data': now, 'fax': now, 'email': now, 'scheduler': now}
        self.modes = ['in', 'out', 'fetched', 'missed', 'in_alt', 'out_alt']
        #self.modes = ['in']
        
    # Return all logged events of type 'type' whose timestamp is newer than the last time we called the function
    # for that type. The type can be 'call', 'sms' 'email' etc. 
    def getNewEvents(self, type):
        # The timestamps in the logs do not behave not as expected..
        # They are adjusted to local time, so if you receive a call right now
        # it will have the time: time.time() + time.altzone
        now = time.time() + time.altzone
        events = []
        for mode in self.modes:
            # usage: log_data_by_time(type, start time, end time, mode]), where mode is 'in', 'out' etc.
            events.extend( logs.log_data_by_time(type, self.last_read_times[type], now, mode) )
        self.last_read_times[type] = now
        return events


#
# Class PhoneListener provides a callback function which logs the current status
# of the phone line.
#
class PhoneListener:  
    def __init__(self, phone_log):
        self.log = phone_log
        self.status = telephone.EStatusIdle
        self.call_number = ""
        self.new_calls = []
    
    def __call__(self, (status, call_number)):
        self.status, self.call_number = (status, call_number)
        print "Phone status changed: " + str(self.getStatusString(self.status))
        # if the status becomes idle it means that we have just completed a call event 
        if self.status == telephone.EStatusIdle:
            self.new_calls.extend(self.log.getNewEvents('call'))
    
    def formattedString(self):
        text = "" + str(self.getStatusString(self.status))
        if self.call_number:
            text += ", Call number: " + str(self.call_number)
        return text
    
    def __str__(self):
        text = "" + str(self.getStatusString(self.status))
        if self.call_number:
            text += ":" + str(self.call_number)
        return text
    
    # Returns a plaintext description of an integer status value
    def getStatusString(self, status):
        s = ""
        if status == telephone.EStatusUnknown:
            s = "Unknown"
        elif status == telephone.EStatusIdle: #Idle line status (no active calls).
            s = "Idle"
        elif status == telephone.EStatusDialling: #Call dialling status.
            s = "Dialling"
        elif status == telephone.EStatusRinging: #Call ringing status.
            s = "Ringing"
        elif status == telephone.EStatusAnswering: #Call answering status.
            s = "Answering"
        elif status == telephone.EStatusConnecting: #Call connecting status.
            s = "Connecting"
        elif status == telephone.EStatusConnected: #Call connected status.
            s = "Connected"
        elif status == telephone.EStatusReconnectPending: #Call is undergoing temporary channel loss and it may or may not be reconnected.
            s = "Reconnect pending"
        elif status == telephone.EStatusDisconnecting: #Call disconnecting status.
            s = "Disconnecting"
        elif status == telephone.EStatusHold: #Call on hold.
            s = "Hold"
        elif status == telephone.EStatusTransferring: #Call is transferring.
            s = "Transferring"
        elif status == telephone.EStatusTransferAlerting: #Call in transfer is alerting the remote party.
            s = "Transfer alerting"
        return s;
    
    def getStatus(self):
        return (self.status, self.call_number)


#
# Provides a callback function to keep track of how many unread messages there are
# in the inbox and when the user reads a new message.
# (The phone log by itself can only tell us when a message arrived, not when it was read.)
#
class MessageListener:
    def __init__(self, folder_type = inbox.EInbox):
        self.folder_type = folder_type
        self.inbox = inbox.Inbox() 
        self.unread_messages = self.getAllUnreadMessages()
        self.newly_read_messages = []
    
    def __call__(self, message_id):
        # when a new sms arrives it is put int the list unread_messages.
        # non-sms messages like bluetooth transfers etc. will also trigger this callback function,
        # but since their message_id is invalid according to the Inbox.unread(message_id) function
        # we will not try to monitor when they are read
        #try:
        #    if self.inbox.unread(message_id):
        self.unread_messages.append(message_id)
        print "New message!"
        #except SymbianError:
        #   # not an sms
        #    print "New file transfer or data message."
    
    def update(self):
        # check if any of the messages listed as unread has now been read
        # and update the lists of unread and newly read messages accordingly
        self.newly_read_messages = []
        for message_id in self.unread_messages:
            try:
                if not self.inbox.unread(message_id):
                    self.unread_messages.remove(message_id)
                    self.newly_read_messages.append(message_id)
            except SymbianError:
                # message_id was not found in the inbox so we remove it.
                # (This may be because the message was deleted before it was
                # read, or because it points to something that is not an SMS,
                # like a file sent over bluetooth.)
                self.unread_messages.remove(message_id)
                #print "Removed non-sms or a deleted sms from unread_messages"
    
    # iterates through the entire inbox and returns a list of all unread messages
    def getAllUnreadMessages(self):
        i = inbox.Inbox()
        messages = i.sms_messages()
        unread_messages = []
        if messages:
            for message_id in messages:
                if i.unread(message_id):
                    unread_messages.append(message_id)
        return unread_messages
        

# "Wrapper" classes for representing phonecalls and messages.
# They provide a nice and unified way for printing them.
class Phonecall:
    #Incoming = 0
    #Outgoing = 1
    
    # Create a new Phonecall object from a logged phonecall event (log_entry) in the phone log.
    # The log_entry is a dictionary consisting of name-value pairs, as returned by functions in the logs module.
    def __init__(self, log_entry):
        self.call = log_entry
        self.connected_timestamp = log_entry['time']
        self.duration = log_entry['duration']
        self.disconnected_timestamp = self.connected_timestamp + self.duration
        #if log_entry['direction'] == u'Incoming':
        #    self.direction = Phonecall.Incoming
        #else:
        #    self.direction = Phonecall.Outgoing
        self.direction = log_entry['direction']
        self.number = log_entry['number']
    
    # Returns a verbose "user readable" string representation of the phonecall.
    def formattedString(self):
        s = ""
        s += "{" + "Connect-time: " + str(time.strftime("%X %x %Z", time.localtime(self.connected_timestamp - time.altzone))) + ", "
        s += "Duration: " + str(self.duration) + ", "
        s += "Direction: " + str(self.direction) + ", "
        s += "Phone number: " + str(self.number) + "}"
        return s
    
    # Returns a concise string representation of the phonecall.
    # This format is used when transmitting data to Betelgeuse.
    def __str__(self):
        return str(self.connected_timestamp) + "," + str(self.duration) + "," + str(self.direction) + "," + str(self.number)

class Message:
    # Construct a new Message object from A) a logged message event in the phone log (log_entry)
    # or B) a message in the phone's inbox (specified by message_id).
    # The log_entry is a dictionary consisting of name-value pairs, as returned by functions in the logs module.
    def __init__(self, log_entry = None, message_id = None):
        # A) take the info from the log_entry
        if log_entry:
            self.message = log_entry
            self.time = log_entry['time']
            #if log_entry['direction'] == u'Incoming':
            #    self.direction = Phonecall.Incoming
            #else:
            #    self.direction = Phonecall.Outgoing
            self.direction = log_entry['direction']
            self.number = log_entry['number']
            self.readtime = None
        # B) take the info from the inbox
        elif message_id:
            i = inbox.Inbox()
            self.time = i.time(message_id)
            matching_log_entry = logs.log_data_by_time('sms', self.time, self.time, 'in')
            if matching_log_entry:
                self.message = matching_log_entry[0]
                self.number = self.message['number']
            else:
                self.message = None
                self.number = "Unknown"
            self.direction = u'Incoming'
            self.readtime = time.time() + time.altzone
        else:
            self.message = None
            self.time = None
            self.direction = None
            self.number = None
            self.readtime = None
    
    # Returns a verbose "user readable" string representation of the message.
    def formattedString(self):
        s = "{" + "Time received: " + str(time.strftime("%X %x %Z", time.localtime(self.connected_timestamp - time.altzone))) + ", "
        s += "Direction: " + str(self.direction) + ", "
        s += "Number: " + str(self.number)
        if self.readtime:
            s += ", Readtime: " + str(self.readtime)
        s += "}"
        return s
    
    # Returns a concise string representation of the message.
    # This format is used when transmitting data to Betelgeuse.
    def __str__(self):
        s = str(self.time) + "," + str(self.direction) + "," + str(self.number)
        if self.readtime:
            s += "," + str(self.readtime)
        return s

def main():
        print "Opening socket to start btg.";
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM);
        try:
            sock.connect(('127.0.0.1', 4321));
        except socket.error, msg:
            if sock:
                sock.close();
            sock = None;
            print "Socket error. " + str(msg);
        sock.close();
        print "Started GSM_Location.";
        gsm_location = Everything();
        while(1):
            try:
                gsm_location.update();
            except:
                print "Unexpected error:", sys.exc_info()[0];
                gsm_location.reset();
            #print gsm_location;
            print gsm_location.formattedString()
            gsm_location.sendGsm();
            gsm_location.sendCalls();
            e32.ao_sleep(30);

if __name__ == "__main__":
    main();
